--------------------------------------------------------------------------------
--	̌vZ
--	́F64bit integer
--	óF64bit ( 32bit,  32bit)
--	xF196clk
--------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity sqrt is
port (
	RST			:in std_logic;							--	1=Reset
	CLK			:in std_logic;							--	CLOCK
	DTEN		:in std_logic;							--	input data enable  1=valid
	DTIN		:in std_logic_vector(63 downto 0);		--	input data 64bit integer
	
	DTOUT    	:out std_logic_vector(63 downto 0); 	--	output data 32.32bit
	DTOUT_VLD	:out std_logic;
	BUSY     	:out std_logic							--	1:calculation
);
end sqrt;

architecture SQRT_64 of sqrt is

	signal start					:std_logic;
	signal delay_cnt				:std_logic_vector(1 downto 0);
	signal shift_en					:std_logic;
	signal not_done_din				:std_logic;
	signal not_done					:std_logic;
	signal not_done_reg				:std_logic;
	signal done						:std_logic;
	signal data_is_greater			:std_logic;
	signal res_plus_sh2				:std_logic_vector(127 downto 0);
	signal res_plus_twice_sh2		:std_logic_vector(127 downto 0);
	signal data_minus_res_plus_sh2	:std_logic_vector(127 downto 0);
	signal data_in					:std_logic_vector(63 downto 0);
	signal internal_data			:std_logic_vector(127 downto 0);
	signal shift_two				:std_logic_vector(127 downto 0);
	signal shift_two_reg			:std_logic_vector(127 downto 0);
	signal result					:std_logic_vector(127 downto 0);
	signal result_din				:std_logic_vector(127 downto 0); 
	signal sqrt64_busy				:std_logic;
	signal sqrt64_data_out			:std_logic_vector(63 downto 0);
	signal d1_done					:std_logic;
	
	constant zero64b				:std_logic_vector(63 downto 0):=X"0000_0000_0000_0000";
	constant zero128b				:std_logic_vector(127 downto 0):=X"0000_0000_0000_0000_0000_0000_0000_0000";

begin

	process(RST,CLK)
	begin
		if RST='1' then
			data_in <= (others=>'0');
		elsif rising_edge(CLK) then
			data_in <= DTIN;
		end if;
	end process;
	
	process(RST,CLK)
	begin
		if RST='1' then
			start <= '0';
		elsif rising_edge(CLK) then
			if DTEN='1' and not_done_reg='0' then
				start <= '1';
			else
				start <= '0';
			end if;
		end if;
	end process;
	
	process(RST,CLK)
	begin
		if RST='1' then
			internal_data <= (others=>'0');
		elsif rising_edge(CLK) then
			if start='1' then
				internal_data <= data_in & zero64b;
			elsif shift_en='1' then
				if not_done='1' and data_is_greater='1' then
					internal_data <= data_minus_res_plus_sh2;
				end if;
			end if;
		end if;
	end process;
	
	
	process(RST,CLK)
	begin
		if RST='1' then
			shift_two <= (others=>'0');
		elsif rising_edge(CLK) then
			if start='1' then
				shift_two(127 downto 126) <= '0' & '1';
				shift_two(125 downto 0)   <= (others=>'0');
			elsif (shift_en = '1') then
				shift_two(127 downto 126) <= '0' & '0';
				shift_two(125 downto 0)   <= shift_two(127 downto 2); 
			end if;
		end if;
	end process;
	
	process(RST,CLK)
	begin
		if RST='1' then
			shift_two_reg <= (others=>'0');
		elsif rising_edge(CLK) then
			shift_two_reg <= shift_two;
		end if;
	end process;
	
	process(RST,CLK)
	begin
		if RST='1' then
			delay_cnt <= (others=>'0');
		elsif rising_edge(CLK) then
			if start='1' or shift_en='1' then
				delay_cnt <= (others=>'0');
			else
				delay_cnt <= delay_cnt + '1';
			end if;
		end if;
	end process;
	
	process(RST,CLK)
	begin
		if RST='1' then
			shift_en <= '0' ;
		elsif rising_edge(CLK) then
			if start='1' then
				shift_en <= '0' ;
			else
				if delay_cnt="01" then
					shift_en <= '1';
				else
					shift_en <= '0';
				end if;
			end if;
		end if;
	end process;
	
	
	process(RST,CLK)
	begin
		if RST='1' then
			not_done <= '0' ;
			not_done_reg <= '0' ;
		elsif rising_edge(CLK) then
			if shift_two=zero128b then
				not_done <= '0';
			else
				not_done <= '1';
			end if;
			not_done_reg <= not_done ;
		end if;
	end process;
	done <= '1' when not_done_reg='1' and not_done='0' else '0';
	
	
	process(RST,CLK)
	begin
		if RST='1' then
			result <= (others=>'0');
		elsif rising_edge(CLK) then
			if start='1' then
				result <= (others=>'0');
			elsif not_done='1' and shift_en='1' then
				result <= '0' & result_din(127 downto 1);
			end if;
		end if;
	end process;
	result_din <= res_plus_twice_sh2 when not_done='1' and data_is_greater='1' else result;
	
	process(RST,CLK)
	begin
		if RST='1' then
			res_plus_sh2 <= (others=>'0');
			res_plus_twice_sh2 <= (others=>'0');
			data_minus_res_plus_sh2 <= (others=>'0');
		elsif rising_edge(CLK) then
			res_plus_sh2 <= result + shift_two;
			res_plus_twice_sh2 <= res_plus_sh2 + shift_two_reg;
		    data_minus_res_plus_sh2 <= internal_data - res_plus_sh2;
		end if;
	end process;
	
	
	process(RST,CLK)
	begin
		if RST='1' then
			data_is_greater <= '0';
		elsif rising_edge(CLK) then
			if internal_data>=res_plus_sh2 then
				data_is_greater <= '1';
			else
				data_is_greater <= '0';
			end if;
		end if;
	end process;
	
	
	process(CLK,RST)
	begin
		if RST='1' then
			sqrt64_data_out <= (others=>'0');
			d1_done <= '0';
		elsif rising_edge(CLK) then
			if done='1' then
				sqrt64_data_out <= result(63 downto 0);
			end if;
			d1_done <= done;
		end if;
	end process;
	DTOUT <= sqrt64_data_out;
	DTOUT_VLD <= d1_done;
	
	
	process(RST,CLK)
	begin
		if RST='1' then
			sqrt64_busy <= '0' ;
		elsif rising_edge(CLK) then
			if done='1' then
				sqrt64_busy <= '0';
			elsif DTEN='1' and not_done_reg='0' then
				sqrt64_busy <= '1';
			end if;
		end if;
	end process;
	BUSY <= sqrt64_busy;

end SQRT_64;
